/*
Paul Jeffris
Life

For PC VGA 320x200 256 colors only....direct screen writes, mode 0x13

Compiles fine in borland, microsoft, or symantec,
in fact an ANSI compiler on a PC I think.

if you use the symantec extended dos model, define _32_BIT for the compile.

cell field is a character array, another array, which has as many rows as
the cell field and 4 columns, is used to store info about first and
last used columns for any particular row in the cell field.


commands:

go(g) - starts the thing going
pause(p) - stops game at any point - can be resumed with go
clear(c) - clears the screen
save file(s) - saves pattern on screen to a file
load file(l) - loads any file saved with above save option
edit(e) - edit the cell field - cursor starts in upper left corner
quit(q) - quit


edit mode...
space - sets cell under cursor on or off
cursor accel(DEL) - toggles number of cells cursor moves for each
                    key push, ranges from 1 to 20 in increments of 5
drawmode(INS) - toggles whether cursor draws(turns cells on) as it
                moves or not.

drawmode and cursor accel can only be toggled while in edit mode.


defining PRETTYCOLORS for the compile makes any dead cell retain its
last color while alive...sort of a cool display.
*/



#include <string.h>
#include <dos.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include "life.h"

#undef toupper /* want the function */


/* global vars */

long Population, Generation;
unsigned char CurrentColor;
unsigned char *InitialPalette;
char *Cells;
int *CellRows;
char _far *pScr;


void _cdecl
main()
{
int i, Going, ch, StillGoing, GotChar;
long length1, length2, clock1, clock2;
char NumBuf[15];
char _far *KeyBufferTail;
Initialize();

GotChar = 0;

KeyBufferTail = (char _far *) 0;

StillGoing = 1;

Going = 0;

while(StillGoing)
  {
  sprintf(NumBuf, "%6ld", Population);
  PrintString(NumBuf, 100,
              __NUMCOLS + __NUMCOLS / 40 + 55, __STATCOLOR);
  while(KeyBufferTail == (char _far *) *((char _far *) 0x0040001C))
    {
    Generation++;
    sprintf(NumBuf, "%6ld", Generation);
    PrintString(NumBuf, 108,
                __NUMCOLS + __NUMCOLS / 40 + 55, __STATCOLOR);
    UpdateCells();
    KillAndSpawn();
    if(!(++CurrentColor))
      CurrentColor = __CELLSTARTCOLOR;
    sprintf(NumBuf, "%6ld", Population);
    PrintString(NumBuf, 100,
                __NUMCOLS + __NUMCOLS / 40 + 55, __STATCOLOR);
    }
  KeyBufferTail = (char _far *) 0;
  if(!GotChar)
    ch = Getch();
  else
    GotChar = 0;
  switch(toupper(ch))
    {
    case 'E':
      CurrentColor = __CELLSTARTCOLOR;
      ch = EnterPattern();
      GotChar = 1;
      Going = 0;
      break;
    case 'G':
      KeyBufferTail = (char _far *) *((char _far *) 0x0040001C);
      Going = 1;
      break;
    case 'Q':
      PrintString("REALLY QUIT? (Y/N)", 84,
                  __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
      if(toupper(Getch()) == 'Y')
        StillGoing = 0;
      else
        PrintString("                  ", 84,
                    __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
      Going = 0;
      break;
    case 'C':
      Going = 0;
      ClearGrid();
      break;
    case 'P':
      Going = 0;
      break;
    case 'F':
      LoadPatternFile();
      Going = 0;
      break;
    case 'S':
      SavePatternFile();
      Going = 0;
      break;
    case 'L':
      LoadPatternFile();
      Going = 0;
      break;
    default:
      if(Going)
        KeyBufferTail = (char _far *) *((char _far *) 0x0040001C);
      break;
    }

  }
CleanUp();
}

void
CleanUp()
{
fcloseall();
WritePalette(InitialPalette);
SetVideoMode(3);
free(Cells);
free(CellRows);
free(InitialPalette);
}

void
Initialize()
{
#ifdef _32_BIT
  pScr = _x386_mk_protected_ptr(0xA0000);
#else
  pScr = (char _far *) 0xA0000000;
#endif
Cells = (char *) calloc(__NUMROWS * __NUMCOLS, 1);
if(Cells == NULL)
  {
  printf("\nCould not allocate memory for life grid...exiting.\n\n");
  exit(-2);
  }
CellRows = (int *) calloc(__NUMROWS << 2, sizeof(int));
if(Cells == NULL)
  {
  printf("\nCould not allocate memory for life grid...exiting.\n\n");
  free(Cells);
  exit(-3);
  }
InitialPalette = (unsigned char *) calloc(768, sizeof(unsigned char));
SetVideoMode(0x13);
ReadPalette(InitialPalette);
if(SetPalette() == __EXIT_FAIL)
  {
  CleanUp();
  printf("\nCould not load palette file....exiting.\n\n");
  exit(-1);
  }
ClearScreen();
PrintString("THE GAME OF LIFE", 2,
            __NUMCOLS + __NUMCOLS / 40, __TITLECOLOR);
PrintString("COMMANDS:",       20,
            __NUMCOLS + __NUMCOLS / 40, __HEADINGCOLOR);
PrintString("GO:           G", 28,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("PAUSE:        P", 36,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("EDIT GRID:    E", 44,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("CLEAR GRID:   C", 52,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("LOAD PATTERN: L", 60,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("SAVE PATTERN: S", 68,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("QUIT PROGRAM: Q", 76,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("POPULATION:", 100,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("0", 100,
            __NUMCOLS + __NUMCOLS / 40 + 75, __STATCOLOR);
PrintString("GENERATION:", 108,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("0", 108,
            __NUMCOLS + __NUMCOLS / 40 + 80, __STATCOLOR);
PrintString("CURSOR ACCEL: ", 176,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("1", 176,
            __NUMCOLS + __NUMCOLS / 80 + 78, __STATCOLOR);
PrintString("AUTODRAW IS ", 184,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("OFF", 184,
            __NUMCOLS + __NUMCOLS / 40 + 70, __STATCOLOR);
PrintString("EDIT CONTROLS:", 120,
            __NUMCOLS + __NUMCOLS / 40, __HEADINGCOLOR);
PrintString("MOVE CURSOR: !#\"$", 128,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("PGUP: %  PGDN: &", 136,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("HOME: \'  END:  *", 144,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("CURSOR ACCEL: DEL", 152,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("TOGGLE DRAW:  INS", 160,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("TOGGLE CELL:  SPC", 168,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
PrintString("POS: ", 192,
            __NUMCOLS + __NUMCOLS / 40, __TEXTCOLOR);
CurrentColor = __CELLSTARTCOLOR;
Generation = 0;
Population = 0;
ClearGrid();
DrawBorder();
}

int
LoadPatternFile()
{
FILE *InFile;
char Header[] = "lifernFile";
char HeadBuf[11];
char FName[13];
int  row, col;
PrintString("FILE NAME: ", 84,
            __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
ReadString(FName, 12, 92, __NUMCOLS + __NUMCOLS / 40, __INPUTCOLOR);
if(strlen(FName) == 0)
  {
  PrintString("                 ", 84,
                         __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
  return __EXIT_FAIL;
  }
InFile = fopen(FName, "rb");
if(InFile == NULL)
  {
  PrintString("NOT FOUND-HIT A KEY", 84,
              __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
  Getch();
  PrintString("                   ", 84,
                         __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
  return __FILE_ERR;
  }
fread((void *)HeadBuf, strlen(Header), 1, InFile);
if(memcmp(HeadBuf, Header, strlen(Header)))
  {
  PrintString("INVALID-HIT A KEY", 84,
              __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
  Getch();
  PrintString("                 ", 84,
                         __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
  return __FILE_ERR;
  }
while(fread((void *) &row, sizeof(int), 1, InFile) &&
      fread((void *) &col, sizeof(int), 1, InFile))
  {
  if((row < __NUMROWS) && (row >= 0) && (col < __NUMCOLS) && (col >= 0))
    {
    *(Cells + (row * __NUMCOLS) + col) = 9;
    *(pScr + __WINDOWTOPEDGE * __SCREENWIDTH + __WINDOWLEFTEDGE +
                     (row * __SCREENWIDTH) + col) = __CELLSTARTCOLOR;
    if(col < *(CellRows + (row << 2)) || *(CellRows + (row << 2)) == -1)
      *(CellRows + (row << 2)) = col;
    if(col > *(CellRows + (row << 2) + 1))
      *(CellRows + (row << 2) + 1) = col;
    Population++;
    }
  }
fclose(InFile);
PrintString("                 ", 84,
            __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
return __RETURN_OK;
}


int
SavePatternFile()
{
int i, j;
FILE *OutFile;
char FName[13];
char Header[] = "lifernFile";
PrintString("FILE NAME: ", 84,
            __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
ReadString(FName, 12, 92, __NUMCOLS + __NUMCOLS / 40, __INPUTCOLOR);
if(strlen(FName) == 0)
  {
  PrintString("                 ", 84,
                         __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
  return __EXIT_FAIL;
  }
OutFile = fopen(FName, "wb");
if(OutFile == NULL)
  return __FILE_ERR;
fwrite((const void *)Header, strlen(Header), 1, OutFile);
for(i = 0 ; i < __NUMROWS ; i++)
  for(j = 0 ; j < __NUMCOLS ; j++)
    {
    if(*(Cells + (i * __NUMCOLS) + j))
      {
      fwrite((const void *)&i, sizeof(int), 1, OutFile);
      fwrite((const void *)&j, sizeof(int), 1, OutFile);
      }
    }
fclose(OutFile);
PrintString("                 ", 84,
            __NUMCOLS + __NUMCOLS / 40, __PROMPTCOLOR);
return __RETURN_OK;
}



void
ReadString(char *Buffer, int Max, int Row, int Col, unsigned char Color)
{
char _far *Scr;
int Length, i, ch;
Scr = pScr + Row * __SCREENWIDTH + Col;
PrintLetter(Scr, '@', Color);
Length = 0;
ch = 0;
while((ch != 13) && (Length < Max))
  {
  if(isalnum(ch) || (ch == '.'))
    {
    *Buffer++ = ch;
    PrintLetter(Scr + Length * 6, ch, Color);
    PrintLetter(Scr + (Length + 1) * 6, '@', Color);
    Length++;
    }
  else if(ch == 8)
    {
    if(Length)
      {
      Length--;
      PrintLetter(Scr + (Length + 1) * 6, ' ', 0);
      PrintLetter(Scr + Length * 6, '@', Color);
      Buffer--;
      *Buffer = '\0';
      }
    }
  ch = Getch();
  ch = toupper(ch);
  }
*Buffer = '\0';
for(i = 0 ; i <= Length ; i++)
  PrintLetter(Scr + (i * 6), ' ', Color);
}


void
DrawBorder()
{
int i;
char _far *Scr1;
char _far *Scr2;
Scr1 = pScr + ((__WINDOWTOPEDGE - 1) * __SCREENWIDTH) + __WINDOWLEFTEDGE - 1;
Scr2 = pScr + (__WINDOWTOPEDGE + __NUMROWS) * __SCREENWIDTH + __WINDOWLEFTEDGE - 1;
for(i = 0; i < __NUMCOLS + 2 ; i++)
  {
  *Scr1++ = __BORDERCOLOR;
  *Scr2++ = __BORDERCOLOR;
  }
Scr1 = pScr + ((__WINDOWTOPEDGE) * __SCREENWIDTH) + __WINDOWLEFTEDGE - 1;
Scr2 = pScr + __WINDOWTOPEDGE * __SCREENWIDTH + __WINDOWLEFTEDGE + __NUMCOLS;
for(i = 0; i < __NUMROWS + 2 ; i++)
  {
  *Scr1 = __BORDERCOLOR;
  *Scr2 = __BORDERCOLOR;
   Scr1 += __SCREENWIDTH;
   Scr2 += __SCREENWIDTH;
  }
}

void
UpdateCells()
{
char *pCells;
int i, j, k;
if((j = *CellRows) != -1)            /* if the row is not empty */
  {
  k = *(CellRows + 1);
  pCells = Cells + j;
  for( ; j <= k ; j++, pCells++)     /* update first row */
    {
    if(*pCells > 8)
      {
      if(!j)
        {
        *(CellRows + 2) = 0;
        (*(pCells + 1))++;
        *(CellRows + 3) = 1;
        (*(pCells + __NUMCOLS))++;
        *(CellRows + 6) = 0;
        (*(pCells + __NUMCOLS + 1))++;
        *(CellRows + 7) = 1;
        }
      else if(j == (__NUMCOLS - 1))
        {
        *(CellRows + 3) = __NUMCOLS - 1;
        (*(pCells - 1))++;
        if((__NUMCOLS - 2) < *(CellRows + 2) || (*(CellRows + 2) == -1))
          *(CellRows + 2) = __NUMCOLS - 2;
        (*(pCells + __NUMCOLS - 1))++;
        if((__NUMCOLS - 2) < *(CellRows + 6))
          *(CellRows + 6) = __NUMCOLS - 2;
        (*(pCells + __NUMCOLS))++;
        *(CellRows + 7) = __NUMCOLS - 1;
        }
      else
        {
        (*(pCells - 1))++;
        if(((j - 1) < *(CellRows + 2)) || (*(CellRows + 2) == -1))
          *(CellRows + 2) = j - 1;
        (*(pCells + 1))++;
        if((j + 1) > *(CellRows + 3))
          *(CellRows + 3) = j + 1;
        (*(pCells + __NUMCOLS - 1))++;
        if(((j - 1) < *(CellRows + 6)) || (*(CellRows + 6) == -1))
          *(CellRows + 6) = j - 1;
        (*(pCells + __NUMCOLS))++;
        (*(pCells + __NUMCOLS + 1))++;
        if((j + 1) > *(CellRows + 7))
          *(CellRows + 7) = j + 1;

        }
      }
    }
  }


if((j = *(CellRows + ((__NUMROWS - 1) << 2))) != -1)  /* update last row */
  {
  k = *(CellRows + ((__NUMROWS - 1) << 2) + 1);
  pCells = Cells + ((__NUMROWS - 1) * __NUMCOLS) + j;
  for( ; j <= k ; j++, pCells++)
    {
    if(*pCells > 8)
      {
      if(!j)
        {
        *(CellRows + ((__NUMROWS - 1) << 2) + 2) = 0;
        (*(pCells + 1))++;
        *(CellRows + ((__NUMROWS - 1) << 2) + 3) = 1;
        (*(pCells - __NUMCOLS))++;
        *(CellRows + ((__NUMROWS - 2) << 2) + 2) = 0;
        (*(pCells - __NUMCOLS + 1))++;
        *(CellRows + ((__NUMROWS - 2) << 2) + 3) = 1;
        }
      else if(j == (__NUMCOLS - 1))
        {
        *(CellRows + ((__NUMROWS - 1) << 2) + 3) = __NUMCOLS - 1;
        (*(pCells - 1))++;
        if(((__NUMCOLS - 2) < *(CellRows + ((__NUMROWS - 1) << 2) + 2)) ||
                          (*(CellRows + ((__NUMROWS - 1) << 2) + 2) == -1))
          *(CellRows + ((__NUMROWS - 1) << 2) + 2) = __NUMCOLS - 2;
        (*(pCells - __NUMCOLS - 1))++;
        if(((__NUMCOLS - 2) < *(CellRows + ((__NUMROWS - 2) << 2) + 2)) ||
                          (*(CellRows + ((__NUMROWS - 2) << 2) + 2) == -1))
          *(CellRows + ((__NUMROWS - 2) << 2) + 2) = __NUMCOLS - 2;
        (*(pCells - __NUMCOLS))++;
        *(CellRows + ((__NUMROWS - 2) << 2) + 3) = __NUMCOLS - 1;
        }
      else
        {
        (*(pCells - 1))++;
        if(((j - 1) < *(CellRows + ((__NUMROWS - 1) << 2) + 2)) ||
                          (*(CellRows + ((__NUMROWS - 1) << 2) + 2) == -1))
          *(CellRows + ((__NUMROWS - 1) << 2) + 2) = j - 1;
        (*(pCells + 1))++;
        if((j + 1) > *(CellRows + ((__NUMROWS - 1) << 2) + 3))
          *(CellRows + ((__NUMROWS - 1) << 2) + 3) = j + 1;
        (*(pCells - __NUMCOLS - 1))++;
        if(((j - 1) < *(CellRows + ((__NUMROWS - 2) << 2) + 2)) ||
                          (*(CellRows + ((__NUMROWS - 2) << 2) + 2) == -1))
          *(CellRows + ((__NUMROWS - 2) << 2) + 2) = j - 1;
        (*(pCells - __NUMCOLS))++;
        (*(pCells - __NUMCOLS + 1))++;
        if((j + 1) > *(CellRows + ((__NUMROWS - 2) << 2) + 3))
          *(CellRows + ((__NUMROWS - 2) << 2) + 3) = j + 1;
        }
      }
    }
  }                                    /* everything else now */
for(i = 1, pCells = Cells + __NUMCOLS ; i < __NUMROWS - 1 ; i++)
  {
  j = *(CellRows + (i << 2));
  if(j != -1)
    {
    k = *(CellRows + (i << 2) + 1);
    if(!j)
      {                                          /* in first column */
      (*(pCells - __NUMCOLS))++;
      *(CellRows + ((i - 1) << 2) + 2) = 0;
      (*(pCells + __NUMCOLS))++;
      *(CellRows + ((i + 1) << 2) + 2) = 0;
      (*(pCells - __NUMCOLS + 1))++;
      if( 1 > *(CellRows + ((i - 1) << 2) + 3))
        *(CellRows + ((i - 1) << 2) + 3) = 1;
      (*(pCells + __NUMCOLS + 1))++;
      if( 1 > *(CellRows + ((i + 1) << 2) + 3))
        *(CellRows + ((i + 1) << 2) + 3) = 1;
      (*(pCells + 1))++;
      if( 1 > *(CellRows + (i << 2) + 3))
        *(CellRows + (i << 2) + 3) = 1;
      j = 1;
      }
    if(k == (__NUMCOLS - 1))                     /* in last column */
      {
      (*(pCells - 2))++;
      if(((__NUMCOLS - 2) < *(CellRows + ((i - 1) << 2) + 2)) ||
                                  (*(CellRows + ((i - 1) << 2) + 2) == -1))
        *(CellRows + ((i - 1) << 2) + 2) = __NUMCOLS - 2;
      (*(pCells + __NUMCOLS - 2))++;
      if(((__NUMCOLS - 2) < *(CellRows + (i << 2) + 2)) ||
                                  (*(CellRows + (i << 2) + 2) == -1))
        *(CellRows + (i << 2) + 2) = __NUMCOLS - 2;
      (*(pCells + 2 * __NUMCOLS - 2))++;
      if(((__NUMCOLS - 2) < *(CellRows + ((i + 1) << 2) + 2)) ||
                                  (*(CellRows + ((i + 1) << 2) + 2) == -1))
        *(CellRows + ((i + 1) << 2) + 2) = __NUMCOLS - 2;
      (*(pCells - 1))++;
      *(CellRows + ((i - 1) << 2) + 3) = __NUMCOLS - 1;
      (*(pCells + 2 * __NUMCOLS - 1))++;
      *(CellRows + ((i + 1) << 2) + 3) = __NUMCOLS - 1;
      k--;
      }
    pCells += j;
    for( ; j <= k ; j++, pCells++)               /* now, really     */
      {                                          /* everything else */
      if(*pCells > 8)
        {
        (*(pCells - __NUMCOLS - 1))++;
        if(((j - 1) < *(CellRows + ((i - 1) << 2) + 2)) ||
                                  (*(CellRows + ((i - 1) << 2) + 2) == -1))
          *(CellRows + ((i - 1) << 2) + 2) = j - 1;
        (*(pCells - 1))++;
        if(((j - 1) < *(CellRows + (i << 2) + 2)) ||
                                  (*(CellRows + (i << 2) + 2) == -1))
          *(CellRows + (i << 2) + 2) = j - 1;
        (*(pCells + __NUMCOLS - 1))++;
        if(((j - 1) < *(CellRows + ((i + 1) << 2) + 2)) ||
                                  (*(CellRows + ((i + 1) << 2) + 2) == -1))
          *(CellRows + ((i + 1) << 2) + 2) = j - 1;
        (*(pCells - __NUMCOLS))++;
        (*(pCells + __NUMCOLS))++;
        (*(pCells - __NUMCOLS + 1))++;
        if((j + 1) > *(CellRows + ((i - 1) << 2) + 3))
          *(CellRows + ((i - 1) << 2) + 3) = j + 1;
        (*(pCells + __NUMCOLS + 1))++;
        if((j + 1) > *(CellRows + ((i + 1) << 2) + 3))
          *(CellRows + ((i + 1) << 2) + 3) = j + 1;
        (*(pCells + 1))++;
        if((j + 1) > *(CellRows + (i << 2) + 3))
          *(CellRows + (i << 2) + 3) = j + 1;
        }
      }
    pCells += __NUMCOLS - k - 1;
    }
  else
    pCells += __NUMCOLS;
  }
}


void
KillAndSpawn()
{
int i, j, k, min, max;
char *EndCount, *EndCount1, *pCells;
char _far *Scr;
Scr = pScr + __WINDOWTOPEDGE * __SCREENWIDTH + __WINDOWLEFTEDGE;
pCells = Cells;
for(i = 0 ; i < __NUMROWS ; i++)
  {
  if((j = *(CellRows + (i << 2) + 2)) != -1)
    {
    k = *(CellRows + (i << 2) + 3);
    *(CellRows + (i << 2)) = -1;
    *(CellRows + (i << 2) + 1) = -1;
    *(CellRows + (i << 2) + 2) = -1;
    *(CellRows + (i << 2) + 3) = -1;
    pCells += j;
    Scr += j;
    min = -1;
    max = 0;
    for( ; j <= k ; j++, pCells++, Scr++)
      {
      if(*pCells)
        {
        if(*pCells == 3)
          {
          *pCells = 9;
          Population++;
          *Scr = CurrentColor;
          max = j;
          if(!(min + 1))
            min = j;
          }
        else if(*pCells < 9)
          *pCells = 0;
        else if(*pCells == 11 || *pCells == 12)
          {
          *pCells = 9;
          max = j;
          if(!(min + 1))
            min = j;
          }
        else
          {
          *pCells = 0;
          Population--;
          *Scr = __GRIDCOLOR
          #if defined PRETTYCOLORS
          + CurrentColor
          #endif
          ;
          }
        }
      }
    if(min + 1)
      {
      *(CellRows + (i << 2)) = min;
      *(CellRows + (i << 2) + 1) = max;
      }
    pCells += __NUMCOLS - k - 1;
    Scr += __SCREENWIDTH - k - 1;
    }
  else
    {
    Scr += __SCREENWIDTH;
    pCells += __NUMCOLS;
    *(CellRows + (i << 2)) = -1;
    *(CellRows + (i << 2) + 1) = -1;
    *(CellRows + (i << 2) + 2) = -1;
    *(CellRows + (i << 2) + 3) = -1;
    }
  }
}


void
ClearGrid()
{
int i, j;
char *EndCount, *EndCount1, *pCells;
Population = 0;
Generation = 0;
EndCount1 = Cells + __NUMCOLS * __NUMROWS;
for(pCells = Cells, i = 0 ; pCells < EndCount1 ; i++)
  {
  EndCount = pCells + __NUMCOLS;
  *(CellRows + (i << 2)) = -1;
  *(CellRows + (i << 2) + 1) = -1;
  *(CellRows + (i << 2) + 2) = -1;
  *(CellRows + (i << 2) + 3) = -1;
  for(j = 0 ; pCells < EndCount ; pCells++, j++)
    {
    *(pScr + ((i + __WINDOWTOPEDGE) * __SCREENWIDTH)
                            + j + __WINDOWLEFTEDGE) = __GRIDCOLOR;
    *pCells = 0;
    }
  }
}


void
ClearScreen()
{
char _far *Scr;
for(Scr = pScr ; Scr < pScr + __SCREENWIDTH * __SCREENHEIGHT ; Scr++)
  *Scr = __BACKGROUNDCOLOR;
}


int
EnterPattern()
{
int LeaveEdit, RetVal, CurrentRow, CurrentCol, ch, i;
static int DrawMode = 0;
static int AccMode = 1;
char *pCells, NumBuf[15];
char _far *Scr;
LeaveEdit = 0;
RetVal = __RETURN_OK;
pCells = Cells;
CurrentRow = 0;
CurrentCol = 0;
Scr = pScr + __WINDOWTOPEDGE * __SCREENWIDTH + __WINDOWLEFTEDGE;
*Scr = CurrentColor;
while(!LeaveEdit)
  {
  sprintf(NumBuf, "%3d, %3d", CurrentRow, CurrentCol);
  PrintString(NumBuf, 192,
            __NUMCOLS + __NUMCOLS / 40 + 20, __STATCOLOR);
  sprintf(NumBuf, "%6ld", Population);
  PrintString(NumBuf, 100,
              __NUMCOLS + __NUMCOLS / 40 + 55, __STATCOLOR);
  if(!GetKey(&ch))
    {
    switch(ch)
      {
      case __UPARROW:
        for(i = 0 ; i < AccMode ; i++)
          if(CurrentRow)
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentRow--;
            Scr -= __SCREENWIDTH;
            *Scr = CurrentColor;
            pCells -= __NUMCOLS;
            }
        break;
      case __DOWNARROW:
        for(i = 0 ; i < AccMode ; i++)
          if(CurrentRow < __NUMROWS - 1)
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentRow++;
            Scr += __SCREENWIDTH;
            *Scr = CurrentColor;
            pCells += __NUMCOLS;
            }
        break;
      case __LEFTARROW:
        for(i = 0 ; i < AccMode ; i++)
          if(CurrentCol)
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentCol--;
            Scr--;
            *Scr = CurrentColor;
            pCells--;
            }
        break;
      case __RIGHTARROW:
        for(i = 0 ; i < AccMode ; i++)
          if(CurrentCol < __NUMCOLS - 1)
            {

            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentCol++;
            Scr++;
            *Scr = CurrentColor;
            pCells++;
            }
        break;
      case __HOME:
        for(i = 0 ; i < AccMode ; i++)
          if(CurrentCol && CurrentRow)
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentCol--;
            CurrentRow--;
            Scr -= (__SCREENWIDTH + 1);
            *Scr = CurrentColor;
            pCells -= (__NUMCOLS + 1);
            }
        break;
      case __PAGEUP:
        for(i = 0 ; i < AccMode ; i++)
          if((CurrentCol < __NUMCOLS - 1) && CurrentRow)
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentCol++;
            CurrentRow--;
            Scr -= (__SCREENWIDTH - 1);
            *Scr = CurrentColor;
            pCells -= (__NUMCOLS - 1);
            }
        break;
      case __END:
        for(i = 0 ; i < AccMode ; i++)
          if(CurrentCol && (CurrentRow < (__NUMROWS - 1)))
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentCol--;
            CurrentRow++;
            Scr += (__SCREENWIDTH - 1);
            *Scr = CurrentColor;
            pCells += (__NUMCOLS - 1);
            }
        break;
      case __PAGEDOWN:
        for(i = 0 ; i < AccMode ; i++)
          if((CurrentCol < __NUMCOLS - 1) && (CurrentRow < (__NUMROWS - 1)))
            {
            if(DrawMode)
              if(!*pCells)
                {
                *pCells = 9;
                Population++;
                if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
                  *(CellRows + (CurrentRow << 2)) = CurrentCol;
                if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
                  *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
                }
            if(!*pCells)
              *Scr = __GRIDCOLOR;
            CurrentCol++;
            CurrentRow++;
            Scr += (__SCREENWIDTH + 1);
            *Scr = CurrentColor;
            pCells += (__NUMCOLS + 1);
            }
        break;
      case __INSERT:
        if(DrawMode)
          {
          PrintString("OFF", 184,
                      __NUMCOLS + __NUMCOLS / 40 + 70, __STATCOLOR);
          DrawMode = 0;
          if(!*pCells)
            {
            *pCells = 9;
            Population++;
            if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                              (*(CellRows + (CurrentRow << 2)) == -1))
              *(CellRows + (CurrentRow << 2)) = CurrentCol;
            if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
              *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;

            }
          }
        else
          {
          PrintString("ON ", 184,
                      __NUMCOLS + __NUMCOLS / 40 + 70, __STATCOLOR);
          DrawMode = 1;
          }
        break;
      case __DELETE:
        if(AccMode == 1)
          {
          AccMode = 4;
          PrintString("4", 176,
                      __NUMCOLS + __NUMCOLS / 40 + 75, __STATCOLOR);
          }
        else
          {
          AccMode = (AccMode + 4) % 23;
          itoa(AccMode, NumBuf, 10);
          if(strlen(NumBuf) == 1)
            strcat(NumBuf, " ");
          PrintString(NumBuf, 176,
                      __NUMCOLS + __NUMCOLS / 40 + 75, __STATCOLOR);
          }

        break;
      default:
        break;
      }
    }
  else
    {
    switch(toupper(ch))
      {
      case __SPACE:
        if(!DrawMode)
          {
          if(!*pCells)
            {
            *pCells = 9;
            Population++;
            if((CurrentCol < *(CellRows + (CurrentRow << 2))) ||
                                  (*(CellRows + (CurrentRow << 2)) == -1))
              *(CellRows + (CurrentRow << 2)) = CurrentCol;
            if(CurrentCol > *(CellRows + (CurrentRow << 2) + 1))
              *(CellRows + (CurrentRow << 2) + 1) = CurrentCol;
            }
          else
            {
            *pCells = 0;
            Population--;
            if(CurrentCol == *(CellRows + (CurrentRow << 2)))
              {
              i = 0;
              while(!(*(pCells + i)) && i < __NUMCOLS)
                i++;
              if(i != __NUMCOLS)
                *(CellRows + (CurrentRow << 2)) = i;
              else
                *(CellRows + (CurrentRow << 2)) = -1;
              }
            if(CurrentCol == *(CellRows + (CurrentRow << 2) + 1))
              {
              i = CurrentCol;
              while(!(*(pCells + i)) && i >= 0)
                i--;
              *(CellRows + (CurrentRow << 2) + 1) = i;
              }
            }
          }
        break;
      case 'G':
        LeaveEdit = 1;
        RetVal = 'G';
        break;
      case 'C':
        ClearGrid();
        *Scr = CurrentColor;
        break;
      case 'Q':
        LeaveEdit = 1;
        RetVal = 'Q';
        break;
      case 'L':
        LeaveEdit = 1;
        RetVal = 'L';
        break;
      case 'S':
        LeaveEdit = 1;
        RetVal = 'S';
        break;
      default:
        break;
      }
    }
  }
if(!*pCells)
  *Scr = __GRIDCOLOR;
PrintString("         ", 192,
            __NUMCOLS + __NUMCOLS / 40 + 20, __STATCOLOR);
return RetVal;
}



int
GetKey(int *OutChar)
{
int status;
union REGS r;
r.h.ah = 0x10;
int86(0x16, &r, &r);
if((r.h.al == 0xE0) || (r.h.al == 0))
  {
  *OutChar = (int) r.h.ah;
  status = 0;
  }
else
  {
  *OutChar = (int) r.h.al;
  status = 1;
  }
return status;
}


int
Getch()
{
union REGS r;
r.h.ah = 0x10;
int86(0x16, &r, &r);
return (int) r.h.al;
}

void
SetVideoMode(char Mode)
{
union REGS r;
r.h.ah = 0;
r.h.al = Mode;
int86(0x10, &r, &r);
}


void
PrintLetter(char _far *Buffer, int x, unsigned char Color)
{
char i, j;
for( i = 0 ; i < 20 ; Buffer += (__SCREENWIDTH - 4))
  {
  for(j = 0  ; j < 4 ; i++, j++, Buffer++)
    {
    if(font[x - 32][i])
      *Buffer = Color;
    else
      *Buffer = __BACKGROUNDCOLOR;
    }
  }
}


void
PrintString(char *String, int Row, int Column, unsigned char Color)
{
int i;
char _far *Scr;
Scr = pScr + Row * __SCREENWIDTH + Column;
for(i = 0 ; i < strlen(String) ; i++)
  {
  PrintLetter(Scr, String[i], Color);
  Scr += 5;
  }
}


int
LoadPaletteFileBinary(char *filename, unsigned char *Palette)
{
FILE *input;
input = fopen(filename, "rb");
if(input == NULL)
  return __EXIT_FAIL;
fread(Palette, 768, 1, input);
fclose(input);
return __RETURN_OK;
}

void
ReadPalette(unsigned char *Palette)
{
char _far *fpPalette;
union REGS r;
struct SREGS sr;
fpPalette = (char _far *)Palette;
sr.es = FP_SEG(fpPalette);
r.x.dx = FP_OFF(fpPalette);
r.x.bx = 0;
r.x.cx = 256;
r.h.ah = 0x10;
r.h.al = 0x17;
int86x(0x10, &r, &r, &sr);
}

int
SetPalette()
{
unsigned char Palette[768];
int i;
unsigned char Red, Green, Blue;
if(LoadPaletteFileBinary("palette.lmp", Palette) == __EXIT_FAIL)
  return __EXIT_FAIL;
WritePalette(Palette);
return __RETURN_OK;
}

void
WritePalette(unsigned char *Palette)
{
char _far *fpPalette;
union REGS r;
struct SREGS sr;
fpPalette = (char _far *)Palette;
sr.es = FP_SEG(fpPalette);
r.x.dx = FP_OFF(fpPalette);
r.x.bx = 0;
r.x.cx = 256;
r.h.ah = 0x10;
r.h.al = 0x12;
int86x(0x10, &r, &r, &sr);
}



/*

#if defined PALETTESTUFF

void
LoadPaletteFileText(char *filename, unsigned char *Palette)
{
char buffer[200], *pbuffer, num1[4], num2[4], num3[4];
int i, j;
int red, green, blue;
FILE *infile;
infile = fopen(filename, "rt");
for(j = 0 ; j < 256 ; j++)
  {
  fgets(buffer, 199, infile);
  pbuffer = buffer;
  i = 0;
  while(*pbuffer && (*pbuffer < '0') && (*pbuffer > '9'))
    pbuffer++;
  while(*pbuffer && (*pbuffer >= '0') && (*pbuffer <= '9'))
    num1[i++] = *pbuffer++;
  num1[i] = '\0';
  i = 0;
  while(*pbuffer && ((*pbuffer < '0') || (*pbuffer > '9')))
    pbuffer++;
  while(*pbuffer && (*pbuffer >= '0') && (*pbuffer <= '9'))
    num2[i++] = *pbuffer++;
  num2[i] = '\0';
  i = 0;
  while(*pbuffer && ((*pbuffer < '0') || (*pbuffer > '9')))
    pbuffer++;
  while(*pbuffer && (*pbuffer >= '0') && (*pbuffer <= '9'))
    num3[i++] = *pbuffer++;
  num3[i] = '\0';
  Palette[3 * j] = (unsigned char) atoi(num1);
  Palette[3 * j + 1] = (unsigned char) atoi(num2);
  Palette[3 * j + 2] = (unsigned char) atoi(num3);
  }
fclose(infile);
}

void
WritePaletteFileText(char *filename, unsigned char *Palette)
{
int i;
FILE *outfile;
outfile = fopen(filename, "wt");
for(i = 0 ; i < 256 ; i++)
  fprintf(outfile, "%d %d %d\n",
                 Palette[3 * i], Palette[3 * i + 1], Palette[3 * i + 2]);
fclose(outfile);
}

void
WritePaletteFileBinary(char *filename, unsigned char *Palette)
{
FILE *outfile;
outfile = fopen(filename, "wb");
fwrite(Palette, 768, 1, outfile);
fclose(outfile);
}



#endif

*/
